"""
Torch dataset support. Any Dataset that implements `TrajectoryDataset` interface can be used for training and evaluation.
"""
from motrack_motion.datasets.torch.core import TorchTrajectoryDataset, TrajectoryDataset
